/*
 *************************************************************************
 *                                                                       *
 * Copyright 2000 Compaq Computer Corporation                            *
 *                                                                       *
 * COMPAQ Registered in U.S. Patent and Trademark Office.                *
 *                                                                       *
 *************************************************************************
 * IMPORTANT: Carefully read the License Terms below before              *
 * proceeding.  By use of these materials you agree to these terms.      *
 * If you do not agree to these terms, you may not use this software or  *
 * the accompanying documentation.                                       *
 *************************************************************************
 * LICENSE TERMS                                                         *
 * 1. GRANT                                                              *
 * Compaq Computer Corporation ("COMPAQ") grants you the right to use,   *
 * modify, and distribute the following source code (the "Software")     *
 * on any number of computers. You may use the Software as part of       *
 * creating a software program or product intended for commercial or     *
 * non-commercial distribution in machine-readable source code, binary,  *
 * or executable formats. You may distribute the Software as             *
 * machine-readable source code provided this license is not removed     *
 * from the Software and any modifications are conspicuously indicated.  *
 * 2. COPYRIGHT                                                          *
 * The Software is owned by COMPAQ and its suppliers and is protected by *
 * copyright laws and international treaties.  Your use of the Software  *
 * and associated documentation is subject to the applicable copyright   *
 * laws and the express rights and restrictions of these terms.          *
 * 3. RESTRICTIONS                                                       *
 * You may not remove any copyright, trademark, or other proprietary     *
 * notices from the Software or the associated  documentation.           *
 * You are responsible for compliance with all applicable export or      *
 * re-export control laws and regulations if you export the Software.    *
 * This license is governed by and is to be construed under the laws     *
 * of the State of Texas.                                                *
 *                                                                       *
 * DISCLAIMER OF WARRANTY AND LIABILITY                                  *
 * Compaq shall not be liable for technical or editorial errors or       *
 * omissions contained herein. The information contained herein is       *
 * subject to change without notice.                                     *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.       *
 * THE ENTIRE RISK ARISING OUT OF THE USE OF THIS SOFTWARE REMAINS WITH  *
 * RECIPIENT.  IN NO EVENT SHALL COMPAQ BE LIABLE FOR ANY DIRECT,        *
 * CONSEQUENTIAL, INCIDENTAL, SPECIAL, PUNITIVE OR OTHER DAMAGES         *
 * WHATSOEVER (INCLUDING WITHOUT LIMITATION DAMAGES FOR LOSS OF BUSINESS *
 * PROFITS, BUSINESS INTERRUPTION, OR LOSS OF BUSINESS INFORMATION),     *
 * EVEN IF COMPAQ HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES    *
 * AND WHETHER IN AN ACTION OF CONTRACT OR TORT INCLUDING NEGLIGENCE.    *
 *                                                                       *
 * If you have any questions concerning this license, please contact:    *
 * Compaq Computer Corporation, Software Business Practices, ZKO1-2/D22, *
 * 110 Spit Brook Road, Nashua, NH. 03062-2698.                          *
 *                                                                       *
 *************************************************************************
 */

/****************************************************************************
**
**  MODULE DESCRIPTION:
**
**      mmap_Jacket & munmap_Jacket
**
**  The reason for a mmap jacket is to handle some VMSisms.
**  One in particular is when it comes to mmaping a rfm=var file.
**  mmaping a rfm=stm (stream-LF) file, works fine.
**  When you attempt to mmap a rfm=var file, you pick up alot of
**  unwanted control characters associated with the record formatting.
**  So in the cases when we attempt to mmap an rfm=var file, we will
**  just read the file into a buffer, and then return that buffer.
**
**  Having done this, we needed to jacket munmap because munmap
**  gets unhappy when you attempt to munmap a buffer instead of a shared
**  memory location (like from real mmap).
**
**
**  MODIFICATION HISTORY:
** 
**  Some callers rely upon the file size information it gets from stat(). Our
**  problem is that for file of "fancy" formats (such as VFC and VAR), the
**  actual number of bytes in the file (which is what stat() reports) is not
**  the same as number of bytes we get to read in from the file (the CRTL
**  removes record headers and inserts LF's to make it look like a stream).
**  So, once we read the file in, and we know that number of bytes read, we
**  need to return this number to our caller. So to achieve this we have a
**  special version of mmap that will return the true number of bytes that
**  have been mapped (via a stat block) and this is GENERIC_MMAP_JACKET8.
**
**  The other argument to GENERIC_MMAP_JACKET8 (usual mmap has only 6 args)
**  is a flags arg. This was added in V7.2 of OpenVMS as an optional 7th
**  arg to mmap.
**
**  Finally, so that we don't duplicate all the mmap code, we turn the
**  "old" mmap jacket into a jacket for our new routine.
*/

#include "vms_jackets.h"
#include <sys/mman.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <stat.h>
#include <stdlib.h>
#include <fabdef.h>

/*
** As of V7.2, mmap can accept an optional 7th argument. Since mmap on pre-7.2
** systems just ignores any extra args, we want to always pass the optional
** 7th arg. So, to keep the compiler happy (since mmap may have been 
** prototyped with either 6 or 7 args) we create a 7 arg prototype and
** use that.
*/

typedef void * (*mmap7)(void *addr, size_t len, int prot, int flags,
		int filedes, off_t off, int gflags);

/***********************************/
/********** munmap_Jacket **********/
/***********************************/
int  GENERIC_MUNMAP_JACKET( void *addr, size_t len )
{
long ptr;
int  using_mmap=1;

if ( getenv(VMS_PREFIX "REALLY_USE_MMAP") )
   return ( munmap( addr, len ) );
else
   {
   /* This is the 2nd part of quadword boundary trick setup in mmap.
   ** If the addr isn't on a quadword boundary then it is 1 of our
   ** buffers which just gets free'd instead of munmap'd.
   */
   ptr = (long)addr;
   using_mmap = ( ptr == (ptr/8)*8 );
   if ( !using_mmap )
      {
      GENERIC_TRACE("munmap_Jacket: free(f00d)","","","");
      free( (char*)addr-1 );
      return 0; /* SUCCESS */
      }
   else
      {
      GENERIC_TRACE("munmap_Jacket: using C RTL munmap","","","");
      return ( munmap( addr, len ) );
      }
   }
}

/*
** According to the documentation, flags must contain one of MAP_FILE or 
** MAP_ANONYMOUS, and one of MAP_VARIABLE or MAP_FIXED. But in the testing
** I've done on an Alpha 7.1 system this doesn't seem to be true. Still,
** since MAP_FILE and MAP_VARIABLE are essentially defaults, we will set them
** in the absense of the opposites.
**
** What is truely different on OpenVMS is that ENXIO is often returned for
** now apparent reason. It turns out that mmap checks the size of the memory
** being mapped against the size of the file, using values from the file's
** header for the comparison. OpenVMS makes no guarantees about *when* file
** updates actually get out to disk (and the file header updated), and quite
** often they don't make it out until the file is closed. This is the problem.
** Even though the file is the correct size, if its just been written the
** file header most likely doesn't reflect the file's true size, and so mmap's
** checks fail. The fix is to force a flush before we call mamp.
*/

/*********************************/
/********** mmap_Jacket8 *********/
/*********************************/
void *GENERIC_MMAP_JACKET8 (void *addr, size_t len, int prot, int flags,
			int filedes, off_t off, stat_t *finfo, int gflags)
{
int VMSflags = flags;
int n=0, read_position=0, extra_bytes=4, buffer_size=0;
char *buft = 0, *extra_bytes_string=0;
mmap7 real_mmap = (mmap7)&mmap;

if (!(VMSflags & MAP_ANONYMOUS))
    VMSflags |= MAP_FILE;
if (!(VMSflags & MAP_FIXED))
    VMSflags |= MAP_VARIABLE;
fsync(filedes);

if ( getenv(VMS_PREFIX "REALLY_USE_MMAP") )
    return real_mmap(addr, len, prot, VMSflags, filedes, off, gflags);
else
    {
    /* sanity checking, if this aint for a file, then we better
    ** use the real mmap.
    */
    if ( !filedes )
       {
       GENERIC_TRACE("mmap_Jacket call w/out a filedes! using C RTL mmap\n","","","");
       return real_mmap (addr, len, prot, VMSflags, filedes, off, gflags);
       }

    GENERIC_TRACE("mmap_Jacket in use, ","","","");

    /*
    ** Even though we are passed in a stat block, we still do an fstat 
    ** ourselves since sometimes not everything has been filled in.
    */
    fstat( filedes, finfo );
    if ( (finfo->st_fab_rfm == FAB$C_FIX) ||
         (finfo->st_fab_rfm == FAB$C_VAR) ||
         (finfo->st_fab_rfm == FAB$C_VFC) )
        {
        GENERIC_TRACE("buffering non-stream file.\n","","","");

        if ( extra_bytes_string = getenv(VMS_PREFIX "MMAP_EXTRA_BYTES") )
            extra_bytes = atoi( extra_bytes_string );

        /* malloc/mmap aligns memory on the quadword boundaries, I am adding
        ** 1 to buft so that when this value eventually gets to munmap, I will
        ** be able to test whether it is an actual mmap return or my buffer
        ** from rfm=var return (mmap actually is page boundary)
        */
	if (finfo->st_fab_rfm == FAB$C_FIX)
	    /* allow for CRTL-inserted LF after each record */
	    extra_bytes += finfo->st_size/finfo->st_fab_mrs;
        buft = (char *)malloc( buffer_size = finfo->st_size+extra_bytes ) +1;
        if ( buft == (char *)1 )
           {
           GENERIC_TRACE("Error! mmap Jacket: malloc failure!\n","","","");
           return (void *)-1; /* failure */
           }

        while(n=read(filedes, buft+read_position, buffer_size-read_position)) {
           if ( n == -1 )
              {
              printf( "\nmmap jacket, some sort of read error.\n" );
              printf( "fid: ( %d, %d, %d )\n", 
			finfo->st_ino[0],finfo->st_ino[1],finfo->st_ino[2] );
              printf( "buffer_size = %d, read_position = %d\n",
                       buffer_size, read_position );
              }
           else
              read_position+=n;
	   if (finfo->st_fab_rfm == FAB$C_VFC) {
	      int actual = n - finfo->st_fab_fsz; /* size minus fixed header */
	      /* move data back to lose fixed header */
	      memmove(buft+read_position-n, /* destination */
		      buft+read_position-actual, /* source */
		      actual); /* size */
              read_position-=finfo->st_fab_fsz;
	   }
	}
        *(buft+read_position) = 0;
	finfo->st_size = read_position;
        return buft;
        }
    else
	{
	void *rv;
        GENERIC_TRACE("using C RTL mmap anyway.","","","");
        rv=real_mmap (addr, len, prot, VMSflags, filedes, off, gflags);
        return rv;
	}
    }
}
/*********************************/
/********** mmap_Jacket **********/
/*********************************/
void *GENERIC_MMAP_JACKET (void *addr, size_t len, int prot, int flags,
			int filedes, off_t off)
{
struct stat finfo;
int gflags = 0;
return GENERIC_MMAP_JACKET8( addr, len, prot, flags, filedes, off, &finfo, gflags);
}

